import os
import json
import requests

from feed_manager.api.models import Feed
from feed_manager.feed_sync.models import ComparableFeedInfo

ORG_FIELD = os.environ.get("ORG_ID_FIELD", "org_key")


class FeedClient(object):
    def __init__(self, api_token, domain, prefix, org_key, add_context):
        self._SCHEME_AUTHORITY = domain
        self.prefix = prefix
        self.api_token = api_token
        self.org_key = org_key
        self.session = requests.Session()
        self.add_context = add_context

        if api_token:
            formatted_api_token = api_token.strip()
            self.session.headers.update({"X-Auth-Token": formatted_api_token})

        # Only needef for the internal calls to the target.
        if org_key and add_context:
            ctx = {"target_org": {ORG_FIELD: org_key}}
            self.session.headers.update({"X-User-Context": json.dumps(ctx)})

    @staticmethod
    def _check_error(resp):
        if resp.status_code < 200 or resp.status_code >= 300:
            print(resp.content)

        resp.raise_for_status()

        if "errorMessage" in resp.json():
            print(resp.json())
            raise requests.exceptions.HTTPError(resp.json()["errorMessage"])

    def get_feed(self, feed_id):
        resp = self.session.get(
            f"{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds/{feed_id}"
        )

        self._check_error(resp)
        return Feed.from_dict(resp.json())

    def get_feeds(self, include_public=False):
        resp = self.session.get(
            f"{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds?include_public={include_public}"
        )

        # Org doesn't exist yet, or has no feeds.
        if resp.status_code == 404:
            return []

        self._check_error(resp)
        feeds = [
            ComparableFeedInfo.from_dict(feed)
            for feed in resp.json()["results"]
            if feed["owner"]
        ]
        return feeds

    def create_feed(self, feed=None, public=False):
        resp = self.session.post(
            f'{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds{"/public" if public else ""}',
            json=feed,
        )

        self._check_error(resp)

        return ComparableFeedInfo.from_dict(resp.json())

    def delete_feed(self, feed_id):
        resp = self.session.delete(
            f"{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds/{feed_id}"
        )

        self._check_error(resp)

        return resp.json()

    def update_feed_info(self, feed_id, feed_info):
        resp = self.session.put(
            f"{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds/{feed_id}/feedinfo",
            json=feed_info,
        )

        self._check_error(resp)

        return resp.json()

    def create_reports(self, feed_id, reports):
        if not reports:
            return {}

        resp = self.session.post(
            f"{self._SCHEME_AUTHORITY}/{self.prefix}/v2/orgs/{self.org_key}/feeds/{feed_id}/reports",
            json={"reports": reports},
        )

        self._check_error(resp)

        return resp.json()
