from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    String,
    Integer,
    ForeignKeyConstraint,
    UniqueConstraint,
    func,
    Index as SqlIndex,
)
from sqlalchemy.ext.declarative import declarative_base


class DictMixin:
    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def as_clear_dict(self):
        _dict = {}
        for c in self.__table__.columns:
            if c.foreign_keys:
                continue
            val = getattr(self, c.name)
            if val:
                _dict[c.name] = val


Base = declarative_base(cls=DictMixin)


class Feed(Base):
    __tablename__ = "feeds"
    id = Column(String, primary_key=True, nullable=False)
    name = Column(String)
    owner = Column(String)
    source_label = Column(String)
    provider_url = Column(String)
    summary = Column(String)
    category = Column(String)
    access = Column(String)
    timestamp = Column(DateTime(timezone=True))
    alertable = Column(Boolean)
    scope = Column(String)
    alert_classification_feature_enabled = Column(Boolean, nullable=True, default=False)

    __table_args__ = (
        UniqueConstraint("id", "owner", "access", name="uc_id_owner_access"),
        UniqueConstraint("owner", "name"),
    )


class Report(Base):
    __tablename__ = "reports"
    id = Column(String, primary_key=True, nullable=False)
    timestamp = Column(DateTime(timezone=True))
    link = Column(String)
    title = Column(String)
    severity = Column(Integer)
    tags = Column(String)
    description = Column(String)
    visibility = Column(String)


class IOC(Base):
    __tablename__ = "iocs"
    id = Column(String, primary_key=True, nullable=False)
    report_id = Column(String, primary_key=True, nullable=False)
    match_type = Column(String)
    field = Column(String)
    values = Column(String)
    link = Column(String)

    __table_args__ = (
        ForeignKeyConstraint(
            (report_id,), (Report.id,), deferrable=True, ondelete="CASCADE"
        ),
        SqlIndex("iocs_report_id idx", report_id),
    )


class ReservedFeedDelegate(Base):
    __tablename__ = "reserved_feed_delegates"
    feed_id = Column(String, primary_key=True, nullable=False)
    org_id = Column(String, primary_key=True, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(
            (feed_id,), (Feed.id,), deferrable=False, ondelete="CASCADE"
        ),
    )


class Watchlist(Base):
    __tablename__ = "watchlists"
    id = Column(String, primary_key=True, nullable=False)
    org_id = Column(String, nullable=False)
    name = Column(String)
    description = Column(String, nullable=True)
    tags_enabled = Column(Boolean, nullable=False)
    alerts_enabled = Column(Boolean, nullable=False)
    alert_classification_enabled = Column(Boolean, nullable=True, default=False)
    create_timestamp = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    last_update_timestamp = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )

    __table_args__ = (UniqueConstraint("name", "org_id"),)


class IgnoreReport(Base):
    __tablename__ = "ignore_reports"
    org_id = Column(String, primary_key=True, nullable=False)
    report_id = Column(String, primary_key=True, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint((report_id,), (Report.id,), ondelete="CASCADE"),
        SqlIndex("ignore_reports_report_id idx", report_id),
    )


class IgnoreReportIOC(Base):
    __tablename__ = "ignore_report_iocs"
    org_id = Column(String, primary_key=True, nullable=False)
    report_id = Column(String, primary_key=True, nullable=False)
    ioc_id = Column(String, primary_key=True, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(
            (
                ioc_id,
                report_id,
            ),
            (
                IOC.id,
                IOC.report_id,
            ),
            ondelete="CASCADE",
        ),
    )


class ReportSeverityOverride(Base):
    __tablename__ = "report_severity_overrides"
    org_id = Column(String, primary_key=True, nullable=False)
    report_id = Column(String, primary_key=True)
    severity = Column(Integer)

    __table_args__ = (
        ForeignKeyConstraint((report_id,), (Report.id,), ondelete="CASCADE"),
        SqlIndex("report_severity_overrides_report_id idx", report_id),
    )


class Classifier(Base):
    __tablename__ = "classifiers"
    id = Column(Integer, primary_key=True)
    key = Column(String)
    value = Column(String)

    __table_args__ = (UniqueConstraint("key", "value", name="uc_key_value_classifier"),)


class ClassifierWatchlist(Base):
    __tablename__ = "classifier_watchlists"
    watchlist_id = Column(String, primary_key=True, nullable=False)
    classifier_id = Column(Integer)

    __table_args__ = (
        ForeignKeyConstraint((watchlist_id,), (Watchlist.id,), ondelete="CASCADE"),
        ForeignKeyConstraint((classifier_id,), (Classifier.id,), ondelete="CASCADE"),
    )


class ReportWatchlist(Base):
    __tablename__ = "report_watchlists"
    watchlist_id = Column(String, primary_key=True, nullable=False)
    report_id = Column(String, primary_key=True, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint((watchlist_id,), (Watchlist.id,), ondelete="CASCADE"),
        ForeignKeyConstraint(
            (report_id,), (Report.id,), deferrable=True, ondelete="CASCADE"
        ),
        SqlIndex("report_watchlists_report_id idx", report_id),
    )


class ReportClassifier(Base):
    __tablename__ = "report_classifiers"
    report_id = Column(String, primary_key=True, nullable=False)
    classifier_id = Column(Integer, primary_key=True, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint((classifier_id,), (Classifier.id,), ondelete="CASCADE"),
        ForeignKeyConstraint(
            (report_id,), (Report.id,), deferrable=True, ondelete="CASCADE"
        ),
        SqlIndex(
            "classifier_report_idx", classifier_id
        ),  # This is actually a covering index on classifier_id that includes report_id
    )


class OrganizationWatchlistState(Base):
    __tablename__ = "organization_watchlist_states"
    org_id = Column(String, primary_key=True, nullable=False)
    dirty_timestamp = Column(DateTime(timezone=True), nullable=False)


class OrganizationWatchlistBuild(Base):
    __tablename__ = "organization_watchlist_builds"
    org_id = Column(String, primary_key=True, nullable=False)
    build_timestamp = Column(DateTime(timezone=True), nullable=False)
