import asyncio
import os
import logging
import sys

from aiohttp import web
from aiohttp.web_routedef import RouteTableDef

from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.engine.watchlist_engine import WatchlistEngine
from feed_manager.api.models import feed_models, watchlist_models
from feed_manager.utils.config import Config
from feed_manager.utils.db import ReadWriteDbPool
from feed_manager.utils.tasks.metric_reporter_task import MetricReporterTask
from feed_manager.utils.tasks.org_watchlist_dirty_task import OrgWatchlistDirtyTask
from feed_manager.utils.tasks.org_watchlist_queue_build_task import (
    OrgWatchlistQueueBuildTask,
)
from feed_manager.utils.tasks.deprovisioning_task import (
    DeprovisioningTask,
    S3DeprovisioningTask,
)
from feed_manager.utils.publisher import Publisher
from feed_manager.utils.tasks.shared_watchlists_task import SharedWatchlistTask
from feed_manager.utils.swagger import swagger_setup, build_definitions
from feed_manager.utils.tasks.update_org_features_task import UpdateOrgFeaturesCacheTask
from feed_manager.utils.tasks.update_public_feed_task import UpdatePublicFeedCacheTask
from feed_manager.web import (
    feedmgr_routes,
    error_handler,
    watchlist_routes,
    watchlist_routes_v2,
    legacy_feeds_routes,
    internal_deprovisioning_routes,
    internal_feedmgr_routes,
    internal_report_routes,
    watchlist_routes_v3,
    feedmgr_routes_v2,
    legacy_doc_routes,
)
from pscr_common.log import setup_logging, clear_logging_handlers, PSCRAccessLogger
from pscr_common.utils import MB
from pscr_common.web.auth import auth_middleware_factory, AuthClient
from pscr_common.web.auditlog import setup_audit_log_config
from pscr_common.metrics import get_reporter
from pscr_common.metadata import MetadataProvider

from aio_pyformance.registry import MetricsRegistry


def build_legacy_feed_app(feed_engine):
    app = web.Application()

    table = legacy_feeds_routes.build_route_table(feed_engine)
    table.setup_routes_and_auth(app)

    return app


def build_feed_app(feed_engine, config):
    app = web.Application()

    table = feedmgr_routes.build_route_table(feed_engine)
    table.setup_routes_and_auth(app)

    table = feedmgr_routes_v2.build_route_table(feed_engine)
    table.setup_routes_and_auth(app)

    swagger_setup(
        app,
        "Feed API",
        swagger_url="/info/api_docs",
        api_base_url=f'{config.get("swagger", "api_prefix")}/feedmgr',
        definitions=build_definitions(feed_models),
    )

    add_legacy_doc_routes(app, config.get("swagger", "api_prefix"), "feedmgr")

    return app


def build_watchlist_app(watchlist_engine, config):
    app = web.Application()

    table = watchlist_routes.build_route_table(watchlist_engine)
    table.setup_routes_and_auth(app)

    table = watchlist_routes_v2.build_route_table(watchlist_engine)
    table.setup_routes_and_auth(app)

    table = watchlist_routes_v3.build_route_table(watchlist_engine)
    table.setup_routes_and_auth(app)

    swagger_setup(
        app,
        "Watchlist API",
        swagger_url="/info/api_docs",
        api_base_url=f'{config.get("swagger", "api_prefix")}/watchlistmgr',
        definitions=build_definitions(watchlist_models),
    )

    add_legacy_doc_routes(app, config.get("swagger", "api_prefix"), "watchlistmgr")

    return app


def build_internal_feed_app(feed_engine):
    app = web.Application()

    table = internal_feedmgr_routes.build_route_table(feed_engine)
    table.setup_routes_and_auth(app)

    return app


def build_internal_report_app(watchlist_engine, feed_engine):
    app = web.Application()

    table = internal_report_routes.build_route_table(watchlist_engine, feed_engine)
    table.setup_routes_and_auth(app)

    return app


def build_internal_deprovisioning_app(feed_engine, watchlist_engine):
    app = web.Application()
    table = internal_deprovisioning_routes.build_route_table(
        feed_engine, watchlist_engine
    )
    table.setup_routes_and_auth(app)
    return app


def build_internal_test_app(main_app):
    internal_test_routes = RouteTableDef()
    for route in main_app.router.routes():
        if hasattr(route, "rbac"):
            # re-registering existing endpoints without setting the rbac property in AuthRouteTableDef
            internal_test_routes.route(route.method, route.resource.canonical)(
                route.handler
            )

    test_app = web.Application()
    test_app.add_routes(internal_test_routes)

    return test_app


def add_legacy_doc_routes(app, api_prefix, url_root):
    table = legacy_doc_routes.build_route_table(api_prefix, url_root)
    table.setup_routes_and_auth(app)


async def create_shared_watchlist_task(app, watchlist_engine):
    shared_watchlist_task = SharedWatchlistTask(watchlist_engine)
    await shared_watchlist_task.register(app)


async def create_watchlist_build_tasks(app, config, db_pool, metrics, watchlist_engine):
    for x in range(config.getint("watchlist", "augmentation_task_count", fallback=1)):
        watchlist_build_task = OrgWatchlistDirtyTask(
            config, db_pool, metrics, watchlist_engine
        )
        await watchlist_build_task.register(app, instance_number=x)

    watchlist_queue_build_task = OrgWatchlistQueueBuildTask(
        config, db_pool, metrics, watchlist_engine
    )
    await watchlist_queue_build_task.register(app)


async def create_metric_reporting_task(app, metrics_registry):
    # Hacky way to check if we are running unit tests
    if "asynctest" in sys.modules.keys():
        return

    env = os.environ.get("ENV_ID", "unknown")
    env_meta = MetadataProvider().get_metadata()
    metrics_reporter = get_reporter(
        metrics_registry,
        env,
        env_meta.task_id,
        "feed_manager",
        env_meta.image.split(":")[-1],
        os.getpid(),
    )
    if metrics_reporter:
        reporter_task = MetricReporterTask(metrics_reporter)
        await reporter_task.register(app)


async def create_update_public_feed_task(app, config, db_pool, feed_engine):
    public_feed_task = UpdatePublicFeedCacheTask(config, db_pool, feed_engine)
    await public_feed_task.register(app)


async def create_update_org_features_task(app, config):
    org_features_task = UpdateOrgFeaturesCacheTask(config)
    await org_features_task.register(app)


async def create_org_deprovisioning_tasks(
    app, config, feed_engine, watchlist_engine, publisher
):
    org_deprovisioning_task = DeprovisioningTask(
        feed_engine, watchlist_engine, publisher
    )
    s3_org_deprovioning_task = S3DeprovisioningTask(config)
    await org_deprovisioning_task.register(app)
    await s3_org_deprovioning_task.register(app)


async def build_app() -> web.Application:
    config = Config.create()

    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=os.environ.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )

    metrics_registry = MetricsRegistry()

    db_pool = ReadWriteDbPool(config)
    db_pool.initialize()

    publisher = Publisher(
        config=config,
        db_pool=db_pool,
        stream_name=os.environ.get("FEEDSEARCHKINESIS_STREAMNAME"),
    )
    feed_engine = FeedEngine(config, db_pool, publisher, metrics_registry)
    report_engine = ReportEngine(config, db_pool, metrics_registry)
    watchlist_engine = WatchlistEngine(
        config,
        db_pool,
        publisher,
        metrics_registry,
        report_engine,
    )

    # Convert max request size megaytes to bytes
    max_request_size_bytes = int(os.environ.get("MAX_REQUEST_SIZE_MB", 250)) * MB

    setup_audit_log_config()

    # RBAC client
    auth_client = AuthClient()

    app = web.Application(
        middlewares=[
            auth_middleware_factory(auth_client),
            error_handler.error_middleware_factory(metrics_registry),
        ],
        client_max_size=max_request_size_bytes,
        handler_args={
            "access_log_class": PSCRAccessLogger,
            "access_log": logging.getLogger("access_log"),
        },
    )

    await auth_client.initialize(app)

    watchlist_app = build_watchlist_app(watchlist_engine, config)
    app.add_subapp("/watchlistmgr/", watchlist_app)
    #    app['watchlistmgr'] = watchlist_app

    feed_app = build_feed_app(feed_engine, config)
    app.add_subapp("/feedmgr/", feed_app)
    #    app['feedmgr'] = feed_app

    legacy_feeds_app = build_legacy_feed_app(feed_engine)
    app.add_subapp("/feeds/", legacy_feeds_app)
    #    app['feeds'] = legacy_feeds_app

    internal_app = build_internal_test_app(app)
    internal_feed_app = build_internal_feed_app(feed_engine)
    internal_deprovisioning_app = build_internal_deprovisioning_app(
        feed_engine, watchlist_engine
    )
    internal_feed_app.add_subapp("/deprovisioning/", internal_deprovisioning_app)
    internal_app.add_subapp("/feedmgr/", internal_feed_app)

    internal_report_app = build_internal_report_app(watchlist_engine, feed_engine)
    internal_app.add_subapp("/reportmgr/", internal_report_app)

    app.add_subapp("/internal/", internal_app)

    await create_watchlist_build_tasks(
        app=app,
        config=config,
        db_pool=db_pool,
        metrics=metrics_registry,
        watchlist_engine=watchlist_engine,
    )

    await create_shared_watchlist_task(app=app, watchlist_engine=watchlist_engine)

    await create_metric_reporting_task(app=app, metrics_registry=metrics_registry)

    await create_update_public_feed_task(
        app=app, config=config, db_pool=db_pool, feed_engine=feed_engine
    )

    await create_update_org_features_task(app=app, config=config)

    await create_org_deprovisioning_tasks(
        app=app,
        config=config,
        feed_engine=feed_engine,
        watchlist_engine=watchlist_engine,
        publisher=publisher,
    )

    # Use this to properly close the TelemetryClient and SearchClient
    app.on_shutdown.append(watchlist_engine.shutdown)
    app.on_shutdown.append(feed_engine.shutdown)

    return app


if __name__ == "__main__":
    webapp = asyncio.get_event_loop().run_until_complete(build_app())

    web.run_app(webapp, port=5000)
