import logging

import rapidjson
from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest

from feed_manager.api.models.feed_models import FeedInfo, Report, Feed
from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.web.route_helpers import (
    get_request_org,
    make_response_json,
    create_feed,
    validate_feed_scope,
)
from feed_manager.web.auth import Resource
from pscr_common.web.auth import AuthRouteTableDef, Action
from pscr_common.web.auditlog import call_audit_log
from pscr_common.org_key import decode_orgkey

_logger = logging.getLogger(__name__)


def build_route_table(engine: FeedEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.post("/v2/orgs/{org_key}/feeds", Resource.feeds, [Action.create])
    async def create_private_feed(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Create a new private feed
        description: Create new private feed.  Unique feed ID will be assigned by the service.  All IOCs will be converted to IOC_V2.  This feed will be owned by the caller. The feed will be available to only the org that created it.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: Feed object
            required: true
            schema:
              $ref: '#/definitions/Feed'
        responses:
            200:
                description: Feed created.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """
        params = await request.json(loads=rapidjson.loads)
        org_key = request.match_info["org_key"]

        rsp = await create_feed(
            engine,
            org_key=org_key,
            params=params,
            required_access=ReportEngine.PRIVATE_LABEL,
        )

        await call_audit_log(
            request,
            description=f"Created new private feed {params['feedinfo']['name']}",
        )
        return rsp

    @routes.post(
        "/v2/orgs/{org_key}/feeds/public", Resource.public_feeds, [Action.create]
    )
    async def create_public_feed(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Create a new public feed
        description: Create public feed.  Unique feed ID will be assigned by the service.  All IOCs will be converted to IOC_V2.  This feed will be owned by the caller.  The feed will be available to all organizations.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: Feed object
            required: true
            schema:
              $ref: '#/definitions/Feed'
        responses:
            200:
                description: Feed created.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """
        params = await request.json(loads=rapidjson.loads)
        org_key = request.match_info["org_key"]

        rsp = await create_feed(
            engine,
            org_key=org_key,
            params=params,
            required_access=ReportEngine.PUBLIC_LABEL,
        )

        await call_audit_log(
            request,
            description=f"Created new public feed {params['feedinfo']['name']}",
        )
        return rsp

    @routes.post("/v2/orgs/{org_key}/feeds/reserved", Resource.feeds, [Action.create])
    async def create_reserved_feed(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Create a new reserved feed
        description: Create reserved feed.  Unique feed ID will be assigned by the service.  This feed will be owned by `org_key` and available to `org_key`'s delegates at the time of creation.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: JSON payload including Feedinfo, Reports list and optional delegates_auth_token.  If the optional delegates_auth_token property is provided it will be used instead of X-Auth-Token header to verify which delegates the caller has access to.  delegates_auth_token has same format as X-Auth-Token - API_SECRET_KEY/API_ID
            required: true
            schema:
              required:
                - feedinfo
                - reports
              properties:
                feedinfo:
                  $ref: '#/definitions/FeedInfo'
                reports:
                  type: array
                  items:
                    $ref: '#/definitions/Report'
                delegates_auth_token:
                  type: string
        responses:
            200:
                description: Feed created.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """

        gateway_url = _get_gateway_url(request)

        params = await request.json(loads=rapidjson.loads)
        auth_token = _get_delegates_auth_token(request, params)

        # Get delegates from app services
        delegates = await engine.delegates_client.get_delegates(
            f"https://{gateway_url}", auth_token
        )

        org_key = request.match_info["org_key"]

        rsp = await create_feed(
            engine,
            org_key=org_key,
            params=params,
            required_access=ReportEngine.RESERVED_LABEL,
            delegates=delegates,
        )

        await call_audit_log(
            request,
            description=f"Created new reserved feed {params['feedinfo']['name']}",
        )
        return rsp

    @routes.get(
        "/v2/orgs/{org_key}/feeds", Resource.feeds, [Action.read], allow_head=False
    )
    async def get_feeds(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get all feeds
        description: Retrieve all feeds owned by the caller.  Provide `include_public=true` parameter to also include public community feeds.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: query
            name: include_public
            description: Include public feeds in listing
            required: false
            type: boolean
        responses:
            200:
                description: Array of feeds
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/Feed'
        """
        org_key = request.match_info["org_key"]

        include_public = request.query.get("include_public", "false").lower() == "true"

        feeds = await engine.get_feeds(org_key, include_public)

        updated_feeds = []
        for feed in feeds:
            updated_feeds.append(FeedEngine.remove_empty_feed_scope(feed))

        return make_response_json({"results": updated_feeds})

    @routes.get(
        "/v2/orgs/{org_key}/feeds/{feed_id}",
        Resource.feeds,
        [Action.read],
        allow_head=False,
    )
    async def get_feed(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get feed
        description: Retrieve feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            200:
                description: Feed
                schema:
                  $ref: '#/definitions/Feed'
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        feed = await engine.get_feed(org_key, feed_id)

        return make_response_json(
            Feed(FeedEngine.remove_empty_feed_scope(feed.feedinfo), feed.reports)
        )

    @routes.delete(
        "/v2/orgs/{org_key}/feeds/{feed_id}", Resource.feeds, [Action.delete]
    )
    async def remove_feed(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Delete feed
        description: Delete feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            204:
                description: Feed deleted
            400:
                description: Unknown feed.
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        feed_name = await engine.get_feed_name(feed_id)

        await engine.remove_feed(org_key, feed_id)

        await call_audit_log(
            request, description=f"Removed feed, '{feed_name}' (ID: {feed_id})"
        )
        return make_response_json()

    @routes.get(
        "/v2/orgs/{org_key}/feeds/{feed_id}/feedinfo",
        Resource.feeds,
        [Action.read],
        allow_head=False,
    )
    async def get_feedinfo(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get feed info
        description: Retrieve feed info metadata for feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            200:
                description: Feed Info
                schema:
                  $ref: '#/definitions/FeedInfo'
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        feedinfo = await engine.get_feedinfo(
            org_key, feed_id, include_reports_count=True
        )

        return make_response_json(FeedEngine.remove_empty_feed_scope(feedinfo))

    @routes.put(
        "/v2/orgs/{org_key}/feeds/{feed_id}/feedinfo", Resource.feeds, [Action.update]
    )
    async def update_feedinfo(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Update feed info
        description: Update feed info metadata for feed with `feed_id`.  This feed must be owned by the caller. Access may not be changed from initial value.
        consumes:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: body
            name: body
            description: FeedInfo object
            required: true
            schema:
              $ref: '#/definitions/FeedInfo'
        produces:
        - application/json
        responses:
            200:
                description: Feed info updated.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: Invalid FeedInfo.
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]
        params = await request.json(loads=rapidjson.loads)

        if "owner" not in params:
            params["owner"] = org_key

        params["id"] = feed_id

        feed_scope = params.get("scope", None)

        if params["access"] != ReportEngine.PUBLIC_LABEL and feed_scope is not None:
            raise HTTPBadRequest(
                reason=f"feedinfo.scope is not applicable for private or reserved feed"
            )

        # validate_feed_scope(feed_scope)

        # cannot change access on a report because we need to gate who can create public feeds
        current_feedinfo = await engine.get_feedinfo(get_request_org(request), feed_id)

        # Existing non scoped feed can not be updated to scoped feed
        if feed_scope is not None and current_feedinfo.scope is None:
            raise HTTPBadRequest(
                reason=f"Currently Updating feed with feed scope is not supported."
            )

        if feed_scope is not None and current_feedinfo.scope is not None:
            if sorted(feed_scope) != sorted(current_feedinfo.scope):
                raise HTTPBadRequest(
                    reason=f"Scope can not be modified on the existing feed."
                )

        if params["access"] != current_feedinfo.access:
            raise HTTPBadRequest(reason="Access does not match current access level")

        # So we don't accidentally overwrite un-alertable status if it was omitted
        if params.get("alertable") is None and current_feedinfo.alertable is False:
            params["alertable"] = False

        feedinfo = FeedInfo.from_dict(params)

        await engine.upsert_feedinfo(org_key, feed_id, FeedInfo.from_dict(params))
        feed_name = feedinfo.name

        await call_audit_log(
            request, description=f"Updated feed, '{feed_name}' (ID: {feed_id})"
        )
        return make_response_json(FeedEngine.remove_empty_feed_scope(feedinfo))

    @routes.get(
        "/v2/orgs/{org_key}/feeds/{feed_id}/reports",
        Resource.feeds,
        [Action.read],
        allow_head=False,
    )
    async def get_reports(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get reports
        description: Retrieve the reports for feed with `feed_id`. Feed must be owned by the caller. Pagination is also supported using `start` and `rows`.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: query
            name: start
            description: 0-based row index from which to begin listing reports
            required: false
            type: int
          - in: query
            name: rows
            description: The number of reports to return starting from `start`
            required: false
            type: int
        responses:
            200:
                description: Reports array
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/Report'
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        skip = max(int(request.query["start"]) if "start" in request.query else 0, 0)
        take = max(int(request.query["rows"]) if "rows" in request.query else 0, 0)

        reports = await engine.get_feed_reports(org_key, feed_id, skip=skip, take=take)

        return make_response_json({"results": reports})

    @routes.get(
        "/v2/orgs/{org_key}/feeds/{feed_id}/reports/{report_id}",
        Resource.feeds,
        [Action.read],
        allow_head=False,
    )
    async def get_report(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get report
        description: Return report with `report_id` for feed.  Feed must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Report
                schema:
                  properties:
                    result:
                      $ref: '#/definitions/Report'
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        report = await engine.get_feed_report(org_key, feed_id, report_id)

        return make_response_json({"report": report})

    @routes.post(
        "/v2/orgs/{org_key}/feeds/{feed_id}/reports", Resource.feeds, [Action.create]
    )
    async def replace_reports(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Replace reports.
        description: Replace reports for feed ID. All IOCs will be converted to IOC_V2. Any existing reports not in the payload will be deleted.  Feed must be owned by the caller.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: body
            name: body
            description: Reports list
            required: true
            schema:
              properties:
                reports:
                  type: array
                  items:
                    $ref: '#/definitions/Report'
        responses:
            200:
                description: success
                schema:
                  type: object
                  required:
                  - success
                  properties:
                    success:
                      type: boolean
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]
        params = await request.json(loads=rapidjson.loads)

        reports = [Report.from_dict(report) for report in params["reports"]]

        inserted, updated, deleted = await engine.replace_reports(
            org_key, feed_id, reports
        )

        reports_info = []
        for report in reports:
            reports_info.append(f"'{report.title}' (ID: {report.id})")

        feed_name = await engine.get_feed_name(feed_id)

        await call_audit_log(
            request,
            description=f"Replaced reports, {', '.join(reports_info)} in feed, '{feed_name}' (ID: {feed_id})",
        )
        return make_response_json({"success": inserted + updated + deleted > 0})

    @routes.put(
        "/v2/orgs/{org_key}/feeds/{feed_id}/reports/{report_id}",
        Resource.feeds,
        [Action.update],
    )
    async def update_report(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Update report (or create, if it doesn't exist)
        description: Update/create report with `report_id` for feed.  All IOCs will be converted to IOC_V2.  Feed must be owned by the caller.
        produces:
        - application/json
        consumes:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: body
            name: body
            description: Report
            required: true
            schema:
              $ref: '#/definitions/Report'
        responses:
            200:
                description: Report
                schema:
                  $ref: '#/definitions/Report'
            400:
                description: Unable to update report because provided timestamp is older than existing timestamp
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        params = await request.json(loads=rapidjson.loads)

        params["id"] = report_id

        report = await engine.upsert_report(org_key, feed_id, Report.from_dict(params))

        report_title = params["title"]
        if report_title is None:
            report_title = report.title
        feed_name = await engine.get_feed_name(feed_id)

        await call_audit_log(
            request,
            description=f"Updated report, '{report_title}' (ID: {report_id}) in feed, '{feed_name}' (ID: {feed_id})",
        )
        return make_response_json(report)

    @routes.delete(
        "/v2/orgs/{org_key}/feeds/{feed_id}/reports/{report_id}",
        Resource.feeds,
        [Action.delete],
    )
    async def remove_report(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Delete report
        description: Delete report with `report_id` for feed .  Feed must be owned by the caller.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            204:
                description: report deleted
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        report_title = await engine.get_report_title(feed_id, report_id)
        feed_name = await engine.get_feed_name(feed_id)

        await engine.delete_report(org_key, feed_id, report_id)

        await call_audit_log(
            request,
            description=f"Deleted report, '{report_title}' (ID: {report_id}) in feed, '{feed_name}' (ID: {feed_id})",
        )
        return make_response_json()

    @routes.get(
        "/v2/orgs/{org_key}/feeds/reserved/{feed_id}/delegates",
        Resource.feeds,
        [Action.read],
    )
    async def get_reserved_feed_delegates(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get delegates for reserved feed
        description: Returns the delegated orgs for feed.  These orgs do not own the feed but have read access to it's contents.  Feed must be owned by `org_key`.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            200:
                description: List of orgs
                schema:
                  properties:
                    delegates:
                      type: array
                      items:
                        type: string
            400:
                description: Feed does not exist or is not a reserved feed
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        delegates = await engine.get_feed_delegates(org_key, feed_id)

        return make_response_json({"delegates": delegates})

    @routes.put(
        "/v2/orgs/{org_key}/feeds/reserved/{feed_id}/delegates",
        Resource.feeds,
        [Action.read],
    )
    async def update_reserved_feed_delegates(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Update reserved orgs for feed
        description: Update the list of organizations that can access this feed.  This uses the appservices delegates API internally.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: body
            name: body
            description: Optional delegates_auth_token property.  If provided it will be used instead of X-Auth-Token header to verify which delegates the caller has access to.  delegates_auth_token has same format as X-Auth-Token - API_SECRET_KEY/API_ID
            required: true
            schema:
              properties:
                delegates_auth_token:
                  type: string
        responses:
            200:
                description: List of delegate orgs
                schema:
                  properties:
                    delegates:
                      type: array
                      items:
                        type: string
            400:
                description: Feed does not exist or is not a reserved feed
        """
        feed_id = request.match_info["feed_id"]
        org_key = request.match_info["org_key"]

        gateway_url = _get_gateway_url(request)

        params = await request.json(loads=rapidjson.loads)
        auth_token = _get_delegates_auth_token(request, params)

        # Get delegates from app services
        delegates = await engine.delegates_client.get_delegates(
            f"https://{gateway_url}", auth_token
        )

        await engine.set_feed_delegates(org_key, feed_id, delegates)

        feed_name = await engine.get_feed_name(feed_id)

        org_ids = []
        for delegate_org_key in delegates:
            decoded = decode_orgkey(orgkey=delegate_org_key)
            org_ids.append(f"{decoded.org_id}")

        await call_audit_log(
            request,
            description=f"Added delegated organizations (ID: {', '.join(org_ids)}) to feed, '{feed_name}' (ID: {feed_id})",
        )
        return make_response_json({"delegates": delegates})

    @routes.post("/v2/query/translate", Resource.feeds, [Action.read])
    async def translate_query(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Convert legacy query
        description: Convert CB EDR query to CBC Enterprise EDR query.  This will adjust field names and other syntax to match CBC Enterprise EDR Solr requirements.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Legacy query
            required: true
            schema:
                type: object
                required:
                - query
                properties:
                  query:
                    type: string
        responses:
            200:
                description: Translated query
                schema:
                  type: object
                  properties:
                    query:
                      type: string
            400:
                description: Unable to convert query due to incompatible fields
        """
        params = await request.json(loads=rapidjson.loads)

        translated_query = await engine.translate_legacy_query(params["query"])

        translated_query = "(" + translated_query + ") -enriched:true"

        return make_response_json({"query": translated_query})

    def _get_gateway_url(request: web.Request):
        request_context = request.headers.get("X-Request-Context")
        if not request_context:
            raise HTTPBadRequest(reason="Missing X-Request-Context")

        request_context = rapidjson.loads(request_context)
        return request_context["gateway_public_url"]

    def _get_delegates_auth_token(request, params):
        auth_token = params.get(
            "delegates_auth_token", request.headers.get("X-Auth-Token")
        )

        if not auth_token:
            raise HTTPBadRequest(
                reason="Missing Auth Token for delegates API. "
                + "Must pass X-Auth-Token header or delegates_auth_token in body."
            )

        return auth_token

    return routes
