import logging
import rapidjson

from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest

from feed_manager.api.models import WatchlistV2, ReportIOCIgnoreList
from feed_manager.engine.watchlist_engine import WatchlistEngine
from feed_manager.web.auth import Resource
from feed_manager.web.route_helpers import get_request_org, make_response_json
from pscr_common.utils import get_utcnow_with_tzinfo
from pscr_common.web.auth import AuthRouteTableDef, Action
from pscr_common.web.auditlog import call_audit_log

_logger = logging.getLogger(__name__)


def build_route_table(engine: WatchlistEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.post("/v2/watchlist", Resource.watchlists, [Action.create])
    async def create_watchlist(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Create new watchlist
        description: Create a new report or classifier watchlist.  Unique watchlist ID will be generated by the service.  Request must specify `report` or `classifier` but not both.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Watchlist payload
            required: true
            schema:
              $ref: '#/definitions/WatchlistV2'
        responses:
            200:
                description: Watchlist created.
                schema:
                  $ref: '#/definitions/WatchlistV2'
            400:
                description: invalid watchlist request.
        """
        now = get_utcnow_with_tzinfo()
        params = await request.json(loads=rapidjson.loads)
        params["id"] = WatchlistEngine.gen_watchlist_id()

        watchlist_request = WatchlistV2.from_dict(params)

        watchlist = await engine.add_watchlist(
            get_request_org(request), watchlist_request, now
        )

        await call_audit_log(
            request,
            description=f"Created new watchlist, '{watchlist.name}' (ID: {watchlist.id}) (v2)",
        )
        return make_response_json(
            WatchlistEngine.remove_empty_alert_classification_field_v2(watchlist)
        )

    @routes.get("/v2/watchlist", Resource.watchlists, [Action.read], allow_head=False)
    async def get_watchlists(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get all watchlists
        description: Retrieve all watchlists owned by the caller.
        produces:
        - application/json
        parameters: []
        responses:
            200:
                description: Array of watchlists
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/WatchlistV2'
        """
        org_id = get_request_org(request)

        watchlists = await engine.get_watchlists(org_id=org_id)

        updated_watchlists = []
        for w in watchlists:
            updated_watchlists.append(
                WatchlistEngine.remove_empty_alert_classification_field_v2(w)
            )

        return make_response_json({"results": updated_watchlists})

    @routes.get(
        "/v2/watchlist/{watchlist_id}",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlist(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get watchlist
        description: Retrieve watchlist with `watchlist_id`
        produces:
        - application/json
        parameters:
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Return watchlist
                schema:
                  $ref: '#/definitions/WatchlistV2'
            404:
                description: Unknown watchlist.
        """
        watchlist_id = request.match_info["watchlist_id"]

        watchlist = (
            await engine.get_watchlists(
                org_id=get_request_org(request), watchlist_id=watchlist_id
            )
        )[0]

        return make_response_json(watchlist)

    @routes.put("/v2/watchlist/{watchlist_id}", Resource.watchlists, [Action.update])
    async def update_watchlist(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Update watchlist
        description: Update watchlist with `watchlist_id`.  This will update the tags and alert status as well as any reports or classifiers attached to the watchlist.  If a field is missing or null (ie tags_enabled) that field will not be updated.  Cannot update report watchlist with empty `report_ids` list.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
          - in: body
            name: body
            description: Watchlist payload
            required: true
            schema:
              $ref: '#/definitions/WatchlistV2'
        responses:
            200:
                description: Return watchlist
                schema:
                  $ref: '#/definitions/WatchlistV2'
            400:
                description: Unknown watchlist or malformed request.
        """
        now = get_utcnow_with_tzinfo()
        params = await request.json(loads=rapidjson.loads)
        watchlist_request = WatchlistV2.from_dict(params)

        watchlist = await engine.update_watchlist(
            get_request_org(request), watchlist_request, now
        )

        await call_audit_log(
            request,
            description=f"Updated watchlist, '{watchlist.name}' (ID: {watchlist.id}) (v2)",
        )
        return make_response_json(
            WatchlistEngine.remove_empty_alert_classification_field_v2(watchlist)
        )

    @routes.get(
        "/v2/ignore/reports/{report_ids}",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_bulk_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Get ignore status for provided report ids
        description: Get current ignore status for report and embedded iocs in provided list of comma-separated `report_ids`.  `report_ids` can be a single id
        produces:
        - application/json
        parameters:
          - in: path
            name: report_ids
            description: comma seperated list of report ids
            required: true
            type: string
        responses:
            200:
                description: Return list of ignored states
                schema:
                  type: array
                  items:
                    $ref: '#/definitions/ReportIOCIgnore'
        """
        report_ids = request.match_info["report_ids"].split(",")
        results = await engine.get_ignores(
            get_request_org(request), report_ids=report_ids
        )

        return make_response_json({"results": results})

    @routes.put("/v2/ignore/reports", Resource.watchlists, [Action.update])
    async def bulk_add_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v2
        summary: Bulk ignore report and iocs
        description: All reports and iocs as defined in the `ReportIOCIgnore` list with ignore=True will not match future events for any watchlist.  All items with ignore=False will enable matching on future events.  A `ReportIOCIgnore` that does not define an `ioc_id` will effect the entire report (all iocs).
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Report IOC Ignore List
            required: true
            schema:
              $ref: '#/definitions/ReportIOCIgnoreList'
        responses:
            200:
                description: Returns ignore status array
                schema:
                  $ref: '#/definitions/ReportIOCIgnoreList'
        """
        params = await request.json(loads=rapidjson.loads)
        ignores_list = ReportIOCIgnoreList.from_dict(params)

        # Raise Bad Request exception if request had MDR reports
        mdr_reports = await engine.get_mdr_reports(get_request_org(request))
        for ignore in ignores_list.ignores:
            if ignore.report_id in mdr_reports:
                raise HTTPBadRequest(
                    reason=f"Report Id - {ignore.report_id} belongs to mdforeedr, "
                    f"therefore can not be ignored."
                )

        ignores = await engine.bulk_update_report_ioc_ignores(
            get_request_org(request), ignores_list.ignores
        )

        ignored_reports_info = []
        for ignore in ignores:
            ignored_ioc_id = ignore.ioc_id
            ignored_report_id = ignore.report_id
            ignored_report_title = await engine.get_report_title(ignored_report_id)
            ignored_reports_info.append(
                f"'{ignored_report_title}' (ID: {ignored_report_id}) and IOC, (ID: {ignored_ioc_id})"
            )

        await call_audit_log(
            request,
            description=f"Ignored reports, {', '.join(ignored_reports_info)} (v2)",
        )
        return make_response_json(ReportIOCIgnoreList(ignores=ignores))

    return routes
